/*************************************************/
/* class to evaluate the rapidity y from x1 & x2 */
/* inherit from LHApdf                           */
/*************************************************/
#include "rapidity.h"

using namespace std;

namespace Moses1 {
// this class inherit functions from the "LHApdf" class

rapidity::rapidity(){}
rapidity::~rapidity(){}



// to return x1 or x2 (colliding quark or antiquark bjorken x)
double rapidity::getBjorkenx(double s, double y, runParameters &runPrm, bool isAnti) { // overload
	double x;
	if (!isAnti) {x = sqrt(s/runPrm.sMax)*exp(y);}
	else {x = sqrt(s/runPrm.sMax)*exp(-y);}
	return(x);
}


// to return the rapidity
double rapidity::getrapidity(double x1, double x2) {
	if (x1>1 || x1<0 || x2>1 ||x2<0) {cerr << "in rapidity.cxx: x1 or x2 is out of range" << endl;}
	return(0.5*log(x1/x2));
}


int rapidity::pdtTOlhaID(int pdtID){
	int lhaID = 0;
	// LHAPDF:
	// index: -6    -5   -4   -3   -2   -1   0   1 2 3 4 5 6
	// parton: tbar bbar cbar sbar ubar dbar glu d u s c b t

	// HEPPDT:
	// particle: 	d u s c b t g
	// id:		1 2 3 4 5 6 21
	switch(pdtID) {
		case PDTGLU: //glu
			lhaID=LHAGLU;
			break;
		default: // all other allowed partons
			if (pdtID >= -6 && pdtID <= 6) {lhaID=pdtID;} // these are the only allowed partons
			else {cerr << "in rapidity.cxx: the supplied pdtID is not supported" << endl;}
			break;
	}
	return(lhaID);
}


// to get ymin and ymax (integration boundaries) depend on s_hat, s_max
double rapidity::getrapidityBoundaries(double s, runParameters &runPrm) {
	if (s<=0) {
		cerr << "warning: in rapidity.cxx: s_hat can not be zero (reset to s_hat_min = 1e-6)" << endl;
		s = 1e-6;
	}
	return(abs(0.5*log(runPrm.sMax/s)));
}



// for this function, remember to call the appropriate constructor !!!
double rapidity::simpsonIntegral(int nMaxSegments, double s, customParticle &parton, runParameters &runPrm, lhapdf &lha) {
	// the Simpson integarting rule:
	// integral(f(x)dx) between [a,b] = (h/3) * { f(x0) + 4f(x1)+2f(x2)+4f(x3)+2f(x4)+...+2f(x[n-2])+4f(x[n-1]) + f(xn) }
	// where: x_n = a + n*h
	// and: h is the step length

	Q = sqrt(s);
	
	nActualSegments = (int)nMaxSegments;

	absBoundary = getrapidityBoundaries(s,runPrm);
	dy = (2*absBoundary)/nActualSegments;
	h = dy/3;
	integrand = 0;
	I=0;
	y = -1*absBoundary; // lower y
	xpdf_q=0;
	xpdf_qbar=0;
	flavor_1 = pdtTOlhaID(parton.pdt_prt.id); // flavor of 'q'
	flavor_2 = -1*flavor_1; // flavor of 'qbar'
		
	int i=0;
	while (y<=absBoundary && i<=nActualSegments) { //i=0(1st point),...,i=nsegments(point nth+1)

		integrand = 0;
		
		// calculate the integrand 
		x_q = getBjorkenx(s, y, runPrm, false); // get x_q(s,y)
		x_qbar = getBjorkenx(s, y, runPrm, true); // get x_qbar(s,y)

		lha.getlhapdf(x_q, Q, flavor_1, xpdf_q); // get x_q*pdf(q)
		lha.getlhapdf(x_qbar, Q, flavor_2, xpdf_qbar); // get x_qbar*pdf(qbar)
		//integrand = ( (xpdf_q/x_q) * (xpdf_qbar/x_qbar) ) / runPrm.sMax;
		integrand = (xpdf_q*xpdf_qbar) * (1./s);

		// simpson
		if (i==0 || i==nActualSegments) {I += integrand;}
		else if (i%2==0 && i!=0 && i!=nActualSegments) {I += 2*integrand;}
		else {I += 4*integrand;}

		// propagate y and i
		y += dy;
		i++;
	}
	return(h*I);
}


// overload to input an external boundary
// for this function, remember to call the appropriate constructor !!!
double rapidity::simpsonIntegral(int nMaxSegments, double s, customParticle &parton, runParameters &runPrm, lhapdf &lha, cutsEmulator &cuts) {
	// the Simpson integarting rule:
	// integral(f(x)dx) between [a,b] = (h/3) * { f(x0) + 4f(x1)+2f(x2)+4f(x3)+2f(x4)+...+2f(x[n-2])+4f(x[n-1]) + f(xn) }
	// where: x_n = a + n*h
	// and: h is the step length
	
	Q = sqrt(s);
	
	nActualSegments = (int)nMaxSegments;
	
	absBoundary = getrapidityBoundaries(s,runPrm);
	if (absBoundary > runPrm.etaCut) {absBoundary = runPrm.etaCut;} //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	

	dy = (2*absBoundary)/nActualSegments;
	h = dy/3;
	integrand = 0;
	I=0;
	y = -1*absBoundary; // lower y
	xpdf_q=0;
	xpdf_qbar=0;
	flavor_1 = pdtTOlhaID(parton.pdt_prt.id); // flavor of 'q'
	flavor_2 = -1*flavor_1; // flavor of 'qbar'
	
	int i=0;
	while (y<=absBoundary && i<=nActualSegments) { //i=0(1st point),...,i=nsegments(point nth+1)
	
		integrand = 0;
	
		// calculate the integrand
		x_q = getBjorkenx(s, y, runPrm, false); // get x_q(s,y)
		x_qbar = getBjorkenx(s, y, runPrm, true); // get x_qbar(s,y)
		lha.getlhapdf(x_q,    Q, flavor_1, xpdf_q); // get x_q*pdf(q)
		lha.getlhapdf(x_qbar, Q, flavor_2, xpdf_qbar); // get x_qbar*pdf(qbar)
		//integrand = ( (xpdf_q/x_q) * (xpdf_qbar/x_qbar) ) / runPrm.sMax;
		integrand = (xpdf_q*xpdf_qbar) * (1./s);
	
		integrand *= cuts.getModifiedCosThetaIntegral(s, y, runPrm); // angular !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		
		// simpson
		if (i==0 || i==nActualSegments) {I += integrand;}
		else if (i%2==0 && i!=0 && i!=nActualSegments) {I += 2*integrand;}
		else {I += 4*integrand;}
	
		// propagate y and i
		y += dy;
		i++;
	}
	return(h*I);
}





// overload to input an external boundary
// for this function, remember to call the appropriate constructor !!!
double rapidity::simpsonIntegral(int nMaxSegments, double s, customParticle &parton, runParameters &runPrm, lhapdf &lha, double cosThetaAbs) {
        // the Simpson integarting rule:
        // integral(f(x)dx) between [a,b] = (h/3) * { f(x0) + 4f(x1)+2f(x2)+4f(x3)+2f(x4)+...+2f(x[n-2])+4f(x[n-1]) + f(xn) }
        // where: x_n = a + n*h
        // and: h is the step length

        Q = sqrt(s);

        nActualSegments = (int)nMaxSegments;

        absBoundary = getrapidityBoundaries(s,runPrm);
        if (absBoundary > (runPrm.etaCut - atanh(cosThetaAbs))) {absBoundary = runPrm.etaCut - atanh(cosThetaAbs);} //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


        dy = (2*absBoundary)/nActualSegments;
        h = dy/3;
        integrand = 0;
        I=0;
        y = -1*absBoundary; // lower y
        xpdf_q=0;
        xpdf_qbar=0;
        flavor_1 = pdtTOlhaID(parton.pdt_prt.id); // flavor of 'q'
        flavor_2 = -1*flavor_1; // flavor of 'qbar'

        int i=0;
        while (y<=absBoundary && i<=nActualSegments) { //i=0(1st point),...,i=nsegments(point nth+1)

                integrand = 0;

                // calculate the integrand
                x_q = getBjorkenx(s, y, runPrm, false); // get x_q(s,y)
                x_qbar = getBjorkenx(s, y, runPrm, true); // get x_qbar(s,y)
                lha.getlhapdf(x_q,    Q, flavor_1, xpdf_q); // get x_q*pdf(q)
                lha.getlhapdf(x_qbar, Q, flavor_2, xpdf_qbar); // get x_qbar*pdf(qbar)
                //integrand = ( (xpdf_q/x_q) * (xpdf_qbar/x_qbar) ) / runPrm.sMax;
                integrand = (xpdf_q*xpdf_qbar) * (1./s);

                // simpson
                if (i==0 || i==nActualSegments) {I += integrand;}
                else if (i%2==0 && i!=0 && i!=nActualSegments) {I += 2*integrand;}
                else {I += 4*integrand;}

                // propagate y and i
                y += dy;
                i++;
        }
        return(h*I);
}







// overload to input an external boundary and forward or backward
// for this function, remember to call the appropriate constructor !!!
double rapidity::simpsonIntegral(int nMaxSegments, double s, customParticle &parton, runParameters &runPrm, lhapdf &lha, cutsEmulator &cuts, double Sq, double Aq, bool isAsymmetric) {
        // the Simpson integarting rule:
        // integral(f(x)dx) between [a,b] = (h/3) * { f(x0) + 4f(x1)+2f(x2)+4f(x3)+2f(x4)+...+2f(x[n-2])+4f(x[n-1]) + f(xn) }
        // where: x_n = a + n*h
        // and: h is the step length
/*
	if (forwardBackward != 1  &&  forwardBackward != -1) {
		cerr << "error: in rapidity.cxx//rapidity::simpsonIntegral()" << endl;
		cerr << "error: invalid forwardBackward value (+1=forward, -1=backward)" << endl;
	}
*/
        Q = sqrt(s);

        nActualSegments = (int)nMaxSegments;

	absBoundary = getrapidityBoundaries(s,runPrm);
	if (absBoundary > runPrm.etaCut) {absBoundary = runPrm.etaCut;} //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
/*	
	if (forwardBackward == -1) {
		//"backward" term (not the full, only the effective part that doesnt vanish in the subtraction f-b)
		y_lowLimit = -1.*absBoundary;
		y_highLimit = 0.;
	}
	else if (forwardBackward == +1) {
		//"forward" term (not the full, only the effective part that doesnt vanish in the subtraction f-b)
		y_lowLimit = 0.;
		y_highLimit = absBoundary;
	}
	dy = absBoundary/(double)nActualSegments; // half of the full range
*/

	y_lowLimit = -1.*absBoundary;
	y_highLimit = absBoundary;
	dy = 2.* absBoundary/(double)nActualSegments;

	h = dy/3.;
	integrand = 0;
	I=0;

	// get the flavours of the quarks
	flavor_1 = pdtTOlhaID(parton.pdt_prt.id); // flavor of 'q'
	flavor_2 = -1*flavor_1; // flavor of 'qbar'

	double sign = 1.;
	if (isAsymmetric) {sign = -1.;}

	double I_forward=0.;
	double I_backward=0.;

	int Sym=+1;
	int Asym=-1;

	double gS=0.;
	double gA=0.;

	double angularS=0.;
	double angularA=0.;
	
	double integrand_forward=0.;
	double integrand_backward=0.;

	y = y_lowLimit;
        int i=0;
	while (y<=y_highLimit && i<=nActualSegments) { //i=0(1st point),...,i=nsegments(point nth+1)

		integrand = 0.;

		// calculate the integrand
		//------------------------------------------------------------------------
		// get Bjorken x's
		x_q = getBjorkenx(s, y, runPrm, false); // get x_q(s,y)
		x_qbar = getBjorkenx(s, y, runPrm, true); // get x_qbar(s,y)
		// PDFs term
		lha.getlhapdf(x_q,    Q, flavor_1, xpdf1); // get x_q*pdf(q)
		lha.getlhapdf(x_qbar, Q, flavor_2, xpdf2); // get x_qbar*pdf(qbar)

		// switch quarks
		lha.getlhapdf(x_q,    Q, flavor_2, xpdf3); // get x_q*pdf(qbar)
		lha.getlhapdf(x_qbar, Q, flavor_1, xpdf4); // get x_qbar*pdf(q)
		//integrand = 2. * (1./(x_q*x_qbar)) * (xpdf1*xpdf2 - xpdf3*xpdf4 ) * (1./runPrm.sMax);	
		//integrand = 2. * (xpdf1*xpdf2 - xpdf3*xpdf4 ) * (1./s);
		//---------------------------------------------------------------------------
		//integrand *= cuts.getModifiedCosThetaIntegral(s, y, runPrm, forwardBackward);
		//---------------------------------------------------------------------------	

		gS = 2. * (xpdf1*xpdf2 + xpdf3*xpdf4 ) * (1./s);
		gA = 2. * (xpdf1*xpdf2 - xpdf3*xpdf4 ) * (1./s);
		angularS = cuts.getModifiedCosThetaIntegral(s, y, runPrm,1, Sym);
		angularA = cuts.getModifiedCosThetaIntegral(s, y, runPrm,1, Asym);

		integrand_forward = 0.;
		integrand_backward = 0.;		

		if (y<0.) {
			integrand_forward  += gS*angularS*Sq - gA*angularA*Aq;
			integrand_backward += gS*angularS*Sq + gA*angularA*Aq;
		}
		else {
			integrand_forward  += gS*angularS*Sq + gA*angularA*Aq;
			integrand_backward += gS*angularS*Sq - gA*angularA*Aq;
		}

		// simpson
		if (i==0 || i==nActualSegments) {
			I_forward += integrand_forward;
			I_backward += integrand_backward;
		}
		else if (i%2==0 && i!=0 && i!=nActualSegments) {
			I_forward += 2.*integrand_forward;
			I_backward += 2.*integrand_backward;
		}
		else {
			I_forward += 4.*integrand_forward;
			I_backward += 4.*integrand_backward;
		}

		// propagate y and i
		y += dy;
		i++;
	}
	I = I_forward + sign*I_backward;
//	cout << "I=" << I << endl;

	return(h*I);
}





// for this function, remember to call the appropriate constructor !!!
double rapidity::simpsonIntegralTest(int nMaxSegments, double Q, lhapdf &lha) {
        // the Simpson integarting rule:
        // integral(f(x)dx) between [a,b] = (h/3) * { f(x0) + 4f(x1)+2f(x2)+4f(x3)+2f(x4)+...+2f(x[n-2])+4f(x[n-1]) + f(xn) }
        // where: x_n = a + n*h
        // and: h is the step length
/*
        // use different number of segments for different Q scale.
        else if (Q>100 && Q<=150) {nActualSegments = (int)(nMaxSegments/5);}
        else if (Q>150) {nActualSegments = (int)nMaxSegments;}
*/
        nActualSegments = (int)nMaxSegments;

        double x=1e-6;
        double dx, xpdf;
        dx = (1.-x)/nActualSegments;
        h = dx/3;

        integrand = 0;
        I=0;
        int i=0;

        while (x<=1 && i<=nActualSegments) {

                // calculate the integrand:
                integrand = 0;
                xpdf=0;

                for (int f=-6 ; f<=6 ; f++) { // loop over all quarks and antiquarks
                        lha.getlhapdf(x, Q, f, xpdf); // get x*pdf(x,Q,f)
                        integrand += xpdf;
                }

                // simpson:
                if (i==0 || i==nActualSegments) {I += integrand;}
                else if (i%2==0 && i!=0 && i!=nActualSegments) {I += 2*integrand;}
                else {I += 4*integrand;}

                // propagate x and i:
                x += dx;
                i++;
        }
        return(h*I);
}


}

