#include "mosesEvent.h"

using namespace std;

namespace Moses1 {
	
mosesEvent::mosesEvent(TFile* file, vector<string>& particleRecordNames)
{
	m_file = file;
	m_particleRecordNames = particleRecordNames;
	
	m_dir = m_file->mkdir("runControl");
    m_dir->cd();
	m_bStatus = true;
	
	fillSupportedParticles();
	fillSupportedJets();
	pickRecords();
	buildControlTree();
	buildSummaryTree();
}

mosesEvent::~mosesEvent(){}

// ########## MUST BE CALLED BY THE USER, AFTER THE CONSTRUCTOR ##########
void mosesEvent::passJetFinderWrappers(mosesClusterJetWrap* clusterJetWrap,
									   mosesCellJetWrap*    cellJetWrap)
{
	string sName = "";
	
	sName = "ClusterJet";
	if(clusterJetWrap!=NULL)
	{
		m_jetRecordNames.push_back(sName);
		m_clusterJetRecord = new mosesClusterJetRecord(sName, m_file);
		m_bRecord.insert( make_pair(sName, true) );
		cout << "initializing " << sName << endl;
	}
	m_clusterJetWrap = clusterJetWrap; // initialize also if =NULL
	
	sName = "CellJet";
	if(cellJetWrap!=NULL)
	{
		m_jetRecordNames.push_back(sName);
		m_cellJetRecord = new mosesCellJetRecord(sName, m_file);
		m_bRecord.insert( make_pair(sName, true) );
		cout << "initializing " << sName << endl;
	}
	m_cellJetWrap = cellJetWrap; // initialize also if =NULL
}
// #######################################################################

void mosesEvent::buildSummaryTree()
{
	m_dir->cd();
    string runHeaderTreeName  = "runHeaderTree";
    string runHeaderTreeTitle = runHeaderTreeName;
	m_runSummaryTree = new TTree(runHeaderTreeName.c_str(), runHeaderTreeTitle.c_str());
	m_beams_id = new vector<int>;
	m_beams_m0 = new vector<double>;
	m_beams_p0 = new vector<double>;
	m_beams_pz = new vector<double>;
	m_sProcessName = new vector<string>;
	m_nProcessCode = new vector<int>;
	m_runSummaryTree->Branch("beamParticle_id", &m_beams_id);
	m_runSummaryTree->Branch("beamParticle_m0", &m_beams_m0);	
	m_runSummaryTree->Branch("beamParticle_p0", &m_beams_p0);
	m_runSummaryTree->Branch("beamParticle_pz", &m_beams_pz);
	m_runSummaryTree->Branch("eCM",          &m_eCM);
	m_runSummaryTree->Branch("nEvents",      &m_nEvents);
	m_runSummaryTree->Branch("sProcessName", &m_sProcessName);
	m_runSummaryTree->Branch("nProcessCode", &m_nProcessCode);
	m_runSummaryTree->Branch("nTried",       &m_nTried);
	m_runSummaryTree->Branch("nSelected",    &m_nSelected);
	m_runSummaryTree->Branch("nAccepted",    &m_nAccepted);
	m_runSummaryTree->Branch("sigma_mb",     &m_sigma_mb);
	m_runSummaryTree->Branch("sigmaErr_mb",  &m_sigmaErr_mb);
	m_runSummaryTree->Branch("mHatMin",      &m_nMHatMin);
	m_runSummaryTree->Branch("mHatMax",      &m_nMHatMax);
}

bool mosesEvent::fillSummaryTree(Pythia& pythia)
{
	// once per run
	m_beams_id->clear();
	m_beams_id->push_back( pythia.info.idA() );
	m_beams_id->push_back( pythia.info.idB() );

	m_beams_m0->clear();
	m_beams_m0->push_back( pythia.info.mA() );
    m_beams_m0->push_back( pythia.info.mB() );

	m_beams_p0->clear();
	m_beams_p0->push_back( pythia.info.eA() );
    m_beams_p0->push_back( pythia.info.eB() );

	m_beams_pz->clear();
	m_beams_pz->push_back( pythia.info.pzA() );
    m_beams_pz->push_back( pythia.info.pzB() );

	m_sProcessName->clear();
	m_sProcessName->push_back( pythia.info.name() );
	m_nProcessCode->clear();
    m_nProcessCode->push_back( pythia.info.code() );

	m_eCM          = pythia.info.eCM();
	m_nEvents      = m_evts;
	m_nTried       = pythia.info.nTried();
	m_nSelected    = pythia.info.nSelected();
	m_nAccepted    = pythia.info.nAccepted();
	m_sigma_mb     = pythia.info.sigmaGen();
	m_sigmaErr_mb  = pythia.info.sigmaErr();

	m_nMHatMin = pythia.settings.parm("PhaseSpace:mHatMin");
	m_nMHatMax = pythia.settings.parm("PhaseSpace:mHatMax");
	
    m_runSummaryTree->Fill();
    return true;
}

bool mosesEvent::writeSummaryTree()
{
    m_dir->cd();
    m_runSummaryTree->Write();
    return true;
}

void mosesEvent::fillSupportedParticles()
{
	//----------------------------------------------------------------------------------------------------
    //----------------------------------------------------------------------------------------------------
	m_supportedParticles.insert( make_pair("muon",     0) );
	m_supportedParticles.insert( make_pair("electron", 1) );
	m_supportedParticles.insert( make_pair("photon",   2) );
	m_supportedParticles.insert( make_pair("neutrino", 3) );
	m_supportedParticles.insert( make_pair("meson",    4) );
	m_supportedParticles.insert( make_pair("baryon",   5) );
	//----------------------------------------------------------------------------------------------------
    //----------------------------------------------------------------------------------------------------
}

void mosesEvent::fillSupportedJets()
{
	//----------------------------------------------------------------------------------------------------
    //----------------------------------------------------------------------------------------------------
	m_supportedJets.insert( make_pair("ClusterJet", 0) );
	m_supportedJets.insert( make_pair("CellJet",    1) );
	//----------------------------------------------------------------------------------------------------
    //----------------------------------------------------------------------------------------------------
}

void mosesEvent::buildControlTree()
{
    m_dir->cd();
    string controlTreeName  = "controlTree";
	string controlTreeTitle = controlTreeName;
    m_controlTree = new TTree(controlTreeName.c_str(),controlTreeTitle.c_str());

	// set the master control tree branches
	for( map<string, int>::iterator ii=m_supportedParticles.begin(); ii!=m_supportedParticles.end(); ++ii)
	{
		string sName = (*ii).first;
		m_controlTree->Branch( ("N"+sName+"s").c_str(), &m_nRecords[ sName ] );	
	}
	
	for( map<string, int>::iterator ii=m_supportedJets.begin(); ii!=m_supportedJets.end(); ++ii)
	{
		string sName = (*ii).first;
		m_controlTree->Branch( ("N"+sName+"s").c_str(), &m_nRecords[ sName ] );	
	}
}


void mosesEvent::pickRecords()
{
	// test
	if((int)m_particleRecordNames.size()==0)
	{
		cout << "error in pickRecords" << endl; 
		exit(-1);
	}

	// initialize
	m_bRecord.clear();
	m_particlesMap.clear();

	// assign names	
    for(int i=0 ; i<(int)m_particleRecordNames.size() ; i++)
    {
		m_bRecord.insert( make_pair(m_particleRecordNames[i], true) );
		m_particlesMap.insert( make_pair(m_particleRecordNames[i], new mosesParticleRecord(m_particleRecordNames[i], m_file)) );
    }
	
	m_hardProcessRecord = new mosesHardProcessRecord("hardProcess", m_file);
}


bool mosesEvent::resetControls()
{
	for( map<string, int>::iterator ii=m_supportedParticles.begin(); ii!=m_supportedParticles.end(); ++ii)
    {
		string sName = (*ii).first;
		m_nRecords[sName] = 0;
    }
	
	for( map<string, int>::iterator ii=m_supportedJets.begin(); ii!=m_supportedJets.end(); ++ii)
    {
		string sName = (*ii).first;
		m_nRecords[sName] = 0;
    }
	
	m_nHardPorcessParticles = 0;
	
	return true;
}


bool mosesEvent::fillControlTree()
{
	m_controlTree->Fill();
	return true;
}


bool mosesEvent::fillAllParticleRecords(Particle &particle, bool isHardProcess)
{
	int nIsHardProcess = (isHardProcess) ? 1 : 0;

	string sName = "";
	bool flag = false;
	//----------------------------------------------------------------------------------------------------
	//----------------------------------------------------------------------------------------------------
	if     (m_bRecord["muon"]     && abs(particle.id())==PDTMU)    { sName = "muon";     flag = true; }
	else if(m_bRecord["electron"] && abs(particle.id())==PDTE)     { sName = "electron"; flag = true; }
	else if(m_bRecord["photon"]   && abs(particle.id())==PDTGAMMA) { sName = "photon";   flag = true; }
	else if(m_bRecord["neutrino"] && ( abs(particle.id())==PDTNUE  ||
									   abs(particle.id())==PDTNUMU ||
									   abs(particle.id())==PDTNUTAU )) { sName = "neutrino"; flag = true; }
	else if(m_bRecord["meson"]    && ( abs(particle.id())>=PDTFIRSTMESON  &&
									   abs(particle.id())<=PDTLASTMESON ))  { sName = "meson"; flag = true; }
	else if(m_bRecord["baryon"]   && ( abs(particle.id())>=PDTFIRSTBARYON &&
									   abs(particle.id())<=PDTLASTBARYON )) { sName = "baryon"; flag = true; }
	//----------------------------------------------------------------------------------------------------
	//----------------------------------------------------------------------------------------------------

	if(flag)
	{
		m_bStatus = m_particlesMap[sName]->fillParticleRecord(particle, nIsHardProcess);
		m_nRecords[sName]++;
	}
	
	return m_bStatus;
}


bool mosesEvent::fillAllJetRecords(Event& event)
{
	string sName = "";
	
	sName = "ClusterJet";
	if(m_bRecord[sName])
	{
		m_bStatus = m_clusterJetWrap->analyzeClusterJet( event );
		if(m_bStatus)
		{
			// assign the jets to hard process participant particles
			m_clusterJetRecord->findHardProcessJets( event, m_clusterJetWrap->getClusterJet() );
			// now pass the wrapper pointer and fill the record:
			m_bStatus = m_clusterJetRecord->fillJetRecord( event, m_clusterJetWrap->getClusterJet() );
			m_nRecords[sName] = m_clusterJetWrap->getClusterJet()->size();
		}
		else cout << sName << "->analyze failed" << endl;
	}
	
	sName = "CellJet";
	if(m_bRecord[sName])
	{
		m_bStatus = m_cellJetWrap->analyzeCellJet( event );
		if(m_bStatus)
		{	// now pass the wrapper pointer and fill the record
			m_bStatus = m_cellJetRecord->fillJetRecord( event, m_cellJetWrap->getCellJet() );
			m_nRecords[sName] = m_cellJetWrap->getCellJet()->size();
		}
		else cout << sName << "->analyze failed" << endl;
	}
	
	return m_bStatus;
}

// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
bool mosesEvent::fillAllRecordTrees() // this needs to change !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
{
	string sName = "";
	
	for( map<string, mosesParticleRecord*>::iterator ii=m_particlesMap.begin() ; ii!=m_particlesMap.end() ; ++ii)
    {
		sName = (*ii).first;
		if(m_bRecord[sName] /*&& m_nRecords[sName]>0*/)
		{
			m_bStatus = m_particlesMap[sName]->fillRecordTree();
			m_bStatus = m_particlesMap[sName]->clearParticleRecord();
		}
		else
		{
			
		}
	}
	
	// jets
	sName = "ClusterJet";
	if(m_bRecord[sName] /*&& m_nRecords[sName]>0*/)
	{
		m_bStatus = m_clusterJetRecord->fillRecordTree();
		m_bStatus = m_clusterJetRecord->clearJetRecord();
	}
	sName = "CellJet";
	if(m_bRecord[sName] /*&& m_nRecords[sName]>0*/)
	{
		m_bStatus = m_cellJetRecord->fillRecordTree();
		m_bStatus = m_cellJetRecord->clearJetRecord();
	}
	
	
	// hard process
	m_hardProcessRecord->fillRecordTree();
	m_hardProcessRecord->clearHardProcessRecord();

	return m_bStatus;
}
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!



bool mosesEvent::clearAllParticleRecords()
{
	string sName = "";

	for( map<string, mosesParticleRecord*>::iterator ii=m_particlesMap.begin(); ii!=m_particlesMap.end(); ++ii)
    {
        sName = (*ii).first;
        if( m_bRecord[sName] )
        {
			m_bStatus = m_particlesMap[sName]->clearParticleRecord();
		}
	}
	
	sName = "ClusterJet";
	if(m_bRecord[sName] && m_nRecords[sName]>0)
	{
		m_bStatus = m_clusterJetRecord->clearJetRecord();
	}
	
	sName = "CellJet";
	if(m_bRecord[sName] && m_nRecords[sName]>0)
	{
		m_bStatus = m_cellJetRecord->clearJetRecord();
	}
	
	m_hardProcessRecord->clearHardProcessRecord();
	
    return m_bStatus;
}




bool mosesEvent::writeAllRecordTrees()
{
	string sName = "";
	for( map<string, mosesParticleRecord*>::iterator ii=m_particlesMap.begin(); ii!=m_particlesMap.end(); ++ii)
	{
		sName = (*ii).first;
		if( m_bRecord[sName] )
        {
			m_bStatus = m_particlesMap[sName]->writeRecordTree();	
		}
	}
	
	sName = "ClusterJet";
	if(m_bRecord[sName] && m_nRecords[sName]>0)
	{
		m_bStatus = m_clusterJetRecord->writeRecordTree();
	}
	
	sName = "CellJet";
	if(m_bRecord[sName] && m_nRecords[sName]>0)
	{
		m_bStatus = m_cellJetRecord->writeRecordTree();
	}
	
	m_hardProcessRecord->writeRecordTree();
	
	return m_bStatus;
}


bool mosesEvent::writeControlTree()
{
	m_dir->cd();
	m_controlTree->Write();
	return true;
}


bool mosesEvent::makeAllTreesFriends()
{
	string sName = "";
	string sPath = "";
	for( map<string, mosesParticleRecord*>::iterator ii=m_particlesMap.begin(); ii!=m_particlesMap.end(); ++ii)
	{
		sName = (*ii).first;
		if(m_bRecord[sName])
        {
			sPath = sName+"/"+sName+"Tree";
			m_controlTree->AddFriend( sPath.c_str(), m_file );
		}
	}
	
	sName = "ClusterJet";
	if(m_bRecord[sName])
	{
		sPath = sName+"/"+sName+"Tree";
		m_controlTree->AddFriend( sPath.c_str(), m_file );
	}
	
	sName = "CellJet";
	if(m_bRecord[sName])
	{
		sPath = sName+"/"+sName+"Tree";
		m_controlTree->AddFriend( sPath.c_str(), m_file );
	}
	
	sPath = "hardProcess/hardProcessTree";
	m_controlTree->AddFriend( sPath.c_str(), m_file );
	
	return true;
}


bool mosesEvent::isHardProcessStatus(int status)
{
	int absStatus = abs(status);
	// Particle::status() = 21 - 29 : particles of the hardest subprocess
	// Particle::status() = 31 - 39 : particles of subsequent subprocesses
	// Particle::status() = 41 - 49 : particles produced by initial-state-showers
	// Particle::status() = 51 - 59 : particles produced by final-state-showers
    if     ( absStatus<=29 && absStatus>=21 ) return true;
	else if( absStatus<=39 && absStatus>=31 ) return true; 
    else if( absStatus<=49 && absStatus>=41 ) return true;
	else if( absStatus<=59 && absStatus>=51 ) return true;
	return false;
}

bool mosesEvent::isHardProcess(Event& event, int index)
{
	if( isHardProcessStatus( event[index].status() ) ) return true;
	
	int momIndex = 0;
	for(int i=0 ; i<(int)event.motherList(index).size() ; i++)
	{
		momIndex = event.motherList(index)[i];
		if( isHardProcessStatus( event[momIndex].status() ) ) return true;
	}

	int sonIndex = 0;
	for(int i=0 ; i<(int)event.daughterList(index).size() ; i++)
	{
		sonIndex = event.daughterList(index)[i];
		if( isHardProcessStatus( event[sonIndex].status() ) ) return true;
	}

	return false;
}



bool mosesEvent::loopOverEvents(int evts, int printMod, Pythia& pythia)
{
	m_evts = evts;
	for(int evt=1 ; evt<=evts ; evt++)
	{
		if(evt%printMod==0) cout << "evt=" << evt << endl;
	
		m_bStatus = resetControls();
	
		m_bStatus = clearAllParticleRecords();

		pythia.next();
		
		//----------------------------------------------------------------------------------------------
		
		if (evt==1)
		{
			string filePath_pythiaEventrecord = "/srv01/tau/hod/moses-1.2.0/data/mosesRunRecord.dat";
			ofstream eventRecord;
			eventRecord.open(filePath_pythiaEventrecord.c_str());
			pythia.event.list(eventRecord);
			eventRecord.close();
		}
		
		//----------------------------------------------------------------------------------------------
		
		//vector<bool> bIndexEvent( (int)pythia.event.size(), false );
		for (int i=0 ; i<(int)pythia.event.size() ; i++)
		{
			bool belongsTohardProcess = isHardProcess(pythia.event, i);
			bool isBeamParticle       = (abs(pythia.event[i].status())==12) ? true:false;

			// fill all the *particles* containers
        	if ( pythia.event[i].isFinal() )
			{
				m_bStatus = fillAllParticleRecords( pythia.event[i], belongsTohardProcess);
			}
			
			// fill the hardProcess container (including *all* particles that participate in the HP, even those that are invisible or not final)
			if(belongsTohardProcess || isBeamParticle)
			{
				m_hardProcessRecord->fillHardProcessRecord( pythia.event[i], i, m_nHardPorcessParticles);
				m_nHardPorcessParticles++;
			}
		}
		
		m_bStatus = fillAllJetRecords( pythia.event );
	
		m_bStatus = fillControlTree();
		
		m_bStatus = fillAllRecordTrees();
	}
	
	m_bStatus = makeAllTreesFriends();

	m_bStatus = writeAllRecordTrees();	

	m_bStatus = writeControlTree();
	
	m_bStatus = fillSummaryTree(pythia);

	m_bStatus = writeSummaryTree();

	m_file->Write();
	
	return m_bStatus;
}

}

